// -*- C++ -*-
/*!
 * @file  rtc_3DShapeControl.h
 * @brief 3D images-shape control
 * @date  $Date$
 *
 * @author EAST JAPAN ACCOUNTING CENTER CO.,LTD.
 *
 * BSD 3-clause "New" or "Revised" License
 *
 * $Id$
 */

#ifndef RTC_3DSHAPECONTROL_H
#define RTC_3DSHAPECONTROL_H

#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/idl/ExtendedDataTypesSkel.h>
#include <rtm/idl/InterfaceDataTypesSkel.h>

// add to code
#include <iostream>
#include <fstream>
#include <string>
#include <stdio.h>
#include <filesystem>
#include <algorithm>

// OpenCV
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#include "local_basic_type.h"
#include "ModeMatrix_3DShapeControl.h"

// Service implementation headers
// <rtc-template block="service_impl_h">

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="consumer_stub_h">
#include "Creation3DResultStatusStub.h"

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="port_stub_h">
// </rtc-template>

using namespace RTC;

#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

//add to code
/*=================================================================================================
  Macro definition
 ================================================================================================*/
#define _DEBUG_SW_

#define IMAGES_LENGTH_2           (2)
#define IMAGES_LENGTH_3           (3)
#define CASE_SYS_MODE             (0)     /**< Case system Mode           */
#define CASE_PLY_MODE             (1)     /**< Case play mode             */

#define SHAPE_OF_3D_INITIAL       (0x00000000u)
#define SHAPE_OF_3D_CREATING      (0x00000001u)
#define SHAPE_OF_3D_CREATED       (0x00000002u)
#define SHAPE_OF_3D_ERROR         (0x0000FFFFu)
#define SHAPE_OF_3D_INIT          (0xFFFFFFFFu)
#define SHAPE_OF_3D_INVALID       (0xFFFFFFFEu)
#define MAKE_SHIFT                (16)
#define MASK_RESULT               (0xFFFF0000u)
#define MASK_SEQNUM               (0x0000FFFFu)

#define FORMAT_COLOR              (24)

#define IMG_POS_L                 (0)
#define IMG_POS_C                 (1)
#define IMG_POS_R                 (2)

#define IMG_TPOS_C                (0)
#define IMG_TPOS_L                (1)
#define IMG_TPOS_R                (1)

#define TIMER_COUNT_INIT          (0)
#define TIMER_COUNT_START         (30)

#define BACKUP_CAP_IMG            (0)
#define BACKUP_CLD_IMG            (1)

#define TYPE_2DCDP1               (0)
#define TYPE_2DCDP2               (1)

/*!
 * @class rtc_3DShapeControl
 * @brief 3D images-shape control
 *
 * Control of the 3-D images-shape system.
 *
 * image data
 *
 * Refer to "functional specification document_3DShapeControl"
 *
 * http://opencv.org/
 *
 */
class rtc_3DShapeControl
  : public RTC::DataFlowComponentBase
{
 public:
  /*!
   * @brief constructor
   * @param manager Maneger Object
   */
  rtc_3DShapeControl(RTC::Manager* manager);

  /*!
   * @brief destructor
   */
  ~rtc_3DShapeControl();

  // <rtc-template block="public_attribute">
  
  // </rtc-template>

  // <rtc-template block="public_operation">
  
  // </rtc-template>

  /***
   * Initialization of the 3DShapeControl RTC
   *
   * The initialize action (on CREATED->ALIVE transition)
   * formaer rtc_init_entry() 
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post ｰ
   * 
   */
   virtual RTC::ReturnCode_t onInitialize();

  /***
   * Finalize of the 3DShapeControl RTC
   *
   * The finalize action (on ALIVE->END transition)
   * formaer rtc_exiting_entry()
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onFinalize();

  /***
   *
   * The startup action when ExecutionContext startup
   * former rtc_starting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStartup(RTC::UniqueId ec_id);

  /***
   *
   * The shutdown action when ExecutionContext stop
   * former rtc_stopping_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onShutdown(RTC::UniqueId ec_id);

  /***
   * Activate the 3DShapeControl RTC
   *
   * The activated action (Active state entry action)
   * former rtc_active_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);

  /***
   * Deactivate the 3DShapeControl RTC
   *
   * The deactivated action (Active state exit action)
   * former rtc_active_exit()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);

  /***
   * Periodic task of the 3DShapeControl RTC
   *
   * The execution action that is invoked periodically
   * former rtc_active_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);

  /***
   * onAborting the 3DShapeControl RTC
   *
   * The aborting action when main logic error occurred.
   * former rtc_aborting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onAborting(RTC::UniqueId ec_id);

  /***
   * onError the 2dcdp RTC
   *
   * The error action in ERROR state
   * former rtc_error_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
  // virtual RTC::ReturnCode_t onError(RTC::UniqueId ec_id);

  /***
   * Reset initialization of the 3DShapeControl RTC
   *
   * The reset action that is invoked resetting
   * This is same but different the former rtc_init_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onReset(RTC::UniqueId ec_id);
  
  /***
   *
   * The state update action that is invoked after onExecute() action
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStateUpdate(RTC::UniqueId ec_id);

  /***
   *
   * The action that is invoked when execution context's rate is changed
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onRateChanged(RTC::UniqueId ec_id);


 protected:
  // <rtc-template block="protected_attribute">
  
  // </rtc-template>

  // <rtc-template block="protected_operation">
  
  // </rtc-template>

  // Configuration variable declaration
  // <rtc-template block="config_declare">

  // </rtc-template>

  // DataInPort declaration
  // <rtc-template block="inport_declare">
  RTC::MultiCameraImages m_img_camera;
  /*!
   * This port inputs the camera image.
   * - Type: RTC::MultiCameraImages
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: Two-cycle per second.
   */
  InPort<RTC::MultiCameraImages> m_img_cameraIn;
  RTC::MultiCameraImages m_img_capture;
  /*!
   * This port inputs the captured image.
   * - Type: RTC::MultiCameraImages
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: Two-cycle per second.
   */
  InPort<RTC::MultiCameraImages> m_img_captureIn;
  RTC::MultiCameraImages m_img_cloud;
  /*!
   * This port inputs the capture image located in the cloud.
   * - Type: RTC::MultiCameraImages
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: Two-cycle per second.
   */
  InPort<RTC::MultiCameraImages> m_img_cloudIn;
  RTC::TimedString m_sys_mode;
  /*!
   * This port inputs the mode information of the 3-D
   * images-shape-system.
   * - Type: RTC::TimedString
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: Two-cycle per second.
   */
  InPort<RTC::TimedString> m_sys_modeIn;
  RTC::TimedString m_play_mode;
  /*!
   * This port inputs a start-information and stop-information of
   * the 3-D images-shape-system.
   * - Type: RTC::TimedString
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: Two-cycle per second.
   */
  InPort<RTC::TimedString> m_play_modeIn;
  
  // </rtc-template>


  // DataOutPort declaration
  // <rtc-template block="outport_declare">
  RTC::MultiCameraImages m_img_2dcdp1;
  /*!
   * This port outputs the left image and the center image.
   * - Type: RTC::MultiCameraImages
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: Two-cycle per second.
   */
  OutPort<RTC::MultiCameraImages> m_img_2dcdp1Out;
  RTC::MultiCameraImages m_img_2dcdp2;
  /*!
   * This port outputs the right image and the center image.
   * - Type: RTC::MultiCameraImages
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: Two-cycle per second.
   */
  OutPort<RTC::MultiCameraImages> m_img_2dcdp2Out;
  
  // </rtc-template>

  // CORBA Port declaration
  // <rtc-template block="corbaport_declare">
  /*!
   * This port used for acquisition of the 3-D image-object
   * creation status.
   * Interface: This port provides for 3-D image-object creation
   *            status from the 3D-Shape RTC.
   */
  RTC::CorbaPort m_Control3DShapeServPort;
  
  // </rtc-template>

  // Service declaration
  // <rtc-template block="service_declare">
  
  // </rtc-template>

  // Consumer declaration
  // <rtc-template block="consumer_declare">
  /*!
   * This port used for acquisition of the 3-D image-object
   * creation status.
   * - Argument:      3-D image-object creation status
   *                  SHAPE_OF_3D_INITIAL
   *                  SHAPE_OF_3D_CREATING
   *                  SHAPE_OF_3D_CREATED
   *                  SHAPE_OF_3D_ERROR
   * - Return Value:  None.
   * - Exception:     None.
   * - PreCondition:  Exists of the 3-D Shape RTC, And connect
   *                  with 3-D Shape RTC.
   * - PostCondition: -
   */
  RTC::CorbaConsumer<Creation3DObjResultStatusServ::Creation3DObjResultStatus> m_Result3DObjStatusCons;
  
  // </rtc-template>

 private:
  // <rtc-template block="private_attribute">
  
  // </rtc-template>

  // <rtc-template block="private_operation">
  
  // </rtc-template>

  // add to macro
  const uword_type IMAGE_SIZE_W_MAX = 640;                        // input cameraimage data maximum width
  const uword_type IMAGE_SIZE_H_MAX = 480;                        // input cameraimage data maximum height
  static const uhword_type MULT_CAMERA_IMG_MAX = 3;               // maximum number of multi-cameraimages

  // add to prototype of func
	void nvd3DShapeCtrlOnInitialize( void );
	void nvd3DShapeCtrlOnActivated( void );
	void nvd3DShapeCtrlOnDeactivated( void );
	void nvdOnExecuteFunc( void );

	void nvdOnExecuteMtx01Matrix( void );
	void nvdOnExecuteMtx02Matrix( void );
	void nvdBackUp3DObjResultNum( uword_type, uword_type );
	void nvd3DShapeCtrlExeOfModeMatrix( uhword_type, uhword_type, uhword_type );
	void nvdCopyToWorkMultiCameraImages( RTC::MultiCameraImages );
	void nvdCopyMultiCameraImages( RTC::MultiCameraImages *, RTC::MultiCameraImages );
	void nvd3D_Image_BackUp( RTC::MultiCameraImages *, RTC::MultiCameraImages, uhword_type );

	bool_type bJudge3DImagesCreation( void );
	bool_type bCopyMultiCameraImages( RTC::MultiCameraImages *, RTC::MultiCameraImages );
	bool_type bCheckMultiCameraImages( RTC::MultiCameraImages );
	bool_type bCheckMultiCameraImagesOfSize( uword_type, uword_type );
	bool_type bCopyTo2dcdpMultiCameraImages( RTC::MultiCameraImages *, RTC::MultiCameraImages *, ubyte_type );

	uword_type  uwGetCreation3DObjResultStatus( void );

	uhword_type uhw3DShapeCtrlInEvCheckofMtx01( void );
	uhword_type uhw3DShapeCtrlInEvCheckofMtx02( void );
	uhword_type uhwEventNumSettingOfMtx01( uhword_type, std::string );
	uhword_type uhwEventNumSettingOfSysMode( std::string );
	uhword_type uhwEventNumSettingOfPlyMode( std::string );
	uhword_type uhwEventNumSettingOfMtx02TimDet( void );
	uhword_type uhwEventNumSettingOfMtx02ImgDet( void );
	uhword_type uhwEventNumSettingOfTimerError( void );
	uhword_type uhwCheck3DObjResultStatus( void );
	uhword_type uhwCheck3DObjResultEndAndErr( uword_type, uword_type );
	uhword_type uhwCheckInputsImageData( void );
	uhword_type uhw3DShapeCtrlGetModeOfMtx01( void );
	uhword_type uhwJudge3D_Image_Creation( RTC::MultiCameraImages );
	uhword_type uhw3D_Image_Create_Retry_Judge( RTC::MultiCameraImages *, uhword_type );

	uhword_type uhwProc_NoStatusChangeMode( void );                 /* No status change of mode                              */
	uhword_type uhwProc_StatusChangeCamera( void );                 /* Status change to camera                               */
	uhword_type uhwProc_StatusChangeCapture( void );                /* Status change to capture                              */
	uhword_type uhwProc_StatusChangeCloud( void );                  /* Status change to cloud                                */
	uhword_type uhwProc_StatusChangeStoredData( void );             /* Status change to stored data                          */
	uhword_type uhwProc_3D_Image_Create_Enable( void );             /* 3D images creation enable by means of mode            */
	uhword_type uhwProc_3D_Image_Create_Disable( void );            /* 3D images creation disable by means of mode           */

	uhword_type uhwProc_3D_Image_Creation_Camera( void );           /* Judge the start of image creation by means of Camera  */
	uhword_type uhwProc_3D_Image_Creation_Capture( void );          /* Judge the start of image creation by means of Capture */
	uhword_type uhwProc_3D_Image_Creation_Cloud( void );            /* Judge the start of image creation by means of Cloud   */
	uhword_type uhwProc_3D_Image_Create_Res_Judge( void );          /* Judge the result of image creation                    */
	uhword_type uhwProc_3D_Image_Create_Err_Judge( void );          /* Judge the error of image creation                     */
	uhword_type uhwProc_3D_Image_BackUp_Capture( void );            /* Back up the image of capture                          */
	uhword_type uhwProc_3D_Image_BackUp_Cloud( void );              /* Back up the image of cloud                            */
	uhword_type uhwProc_NoProc( void );                             /* Nothing                                               */

  // add to variables
	struct validity_st
	{
		uhword_type cap_bkup;
		uhword_type cld_bkup;
	};

	validity_st m_validity;
	RTC::MultiCameraImages m_img_work;
	RTC::MultiCameraImages m_img_cap_bkup;
	RTC::MultiCameraImages m_img_cld_bkup;
	uhword_type m_uhwInEventMtx01;
	uhword_type m_uhwInEventMtx02;
	uhword_type m_uhwMatrixModeMtx01;
	uhword_type m_uhwMatrixModeMtx02;
	uhword_type	m_uhw3DObjMakeErrCnt;
	uword_type  m_uw3DObjResult;
	uword_type  m_uw3DObjResultNum;
	uhword_type m_uhwEnableFlag;

};


extern "C"
{
  DLL_EXPORT void rtc_3DShapeControlInit(RTC::Manager* manager);
};

#endif // RTC_3DSHAPECONTROL_H
